from typing import Type

from gspread import authorize

from nexium_google.spreadsheet.base_row_model import BaseRowModel
from nexium_google.utils import ServiceAccount


SCOPES = ['https://www.googleapis.com/auth/spreadsheets', 'https://www.googleapis.com/auth/drive']


class Spreadsheet:
    def __init__(self, service_account: ServiceAccount, id_: str, models: list[Type[BaseRowModel]]):
        self.client = authorize(
            credentials=service_account.get(
                scopes=SCOPES,
            ),
        )
        self.spreadsheet = self.client.open_by_key(id_)
        self.models = models
        self._initialize_sheets()

    def _initialize_sheets(self):
        for model in self.models:
            # noinspection PyUnresolvedReferences,PyProtectedMember
            if not model._sheet_name:
                raise ValueError(f'Model "{model.__name__}" does not define "__sheet__".')
            # noinspection PyUnresolvedReferences,PyProtectedMember
            sheet = self.spreadsheet.worksheet(model._sheet_name)
            model.bind_sheet(sheet)



from datetime import datetime, date, time

from nexium_google.spreadsheet.field_type import FieldType


class Field:
    def __init__(
            self,
            title: str,
            type_: FieldType,
            nullable: bool = True,
            default: str | int | float | datetime | date | time = None,
            format_: str = None,
    ):
        if not format_ and type_ == FieldType.DATETIME:
            format_ = '%Y-%m-%d %H:%M:%S'
        elif not format_ and type_ == FieldType.DATE:
            format_ = '%Y-%m-%d'
        elif not format_ and type_ == FieldType.TIME:
            format_ = '%H:%M:%S'

        self.title = title
        self.type = type_
        self.nullable = nullable
        self.default = default
        self.format = format_


from json import load

# noinspection PyPackageRequirements
from google.oauth2.service_account import Credentials

from nexium_google.utils.exceptions.service_account import OneFieldRequiredException


class ServiceAccount:
    def __init__(self, json: dict = None, filename: str = None):
        if json:
            self.info = json
        elif filename:
            with open(filename, 'r', encoding='utf-8') as file:
                self.info = load(file)
        else:
            OneFieldRequiredException('One of the fields is required: json, filename')

    def get(self, scopes: list):
        return Credentials.from_service_account_info(
            info=self.info,
            scopes=scopes,
        )



from asyncio import get_running_loop
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime, date, time
from functools import partial

from gspread import Worksheet
from gspread.utils import ValueInputOption

from nexium_google.spreadsheet.field import Field
from nexium_google.spreadsheet.field_type import FieldType


TYPES = {
    FieldType.BOOLEAN: bool,
    FieldType.STRING: str,
    FieldType.INTEGER: int,
    FieldType.FLOAT: float,
    FieldType.DATETIME: datetime,
    FieldType.DATE: date,
    FieldType.TIME: time,
}


class ModelMeta(type):
    def __new__(cls, name, bases, dct):
        fields = {k: v for k, v in dct.items() if isinstance(v, Field)}
        sheet_name = dct.get('__sheet__', None)
        cls_obj = super().__new__(cls, name, bases, dct)
        setattr(cls_obj, '_fields', fields)
        setattr(cls_obj, '_sheet_name', sheet_name)
        return cls_obj


class BaseRowModel(metaclass=ModelMeta):
    _sheet: Worksheet
    _fields: dict[str, Field]

    def __init__(self, **kwargs):
        self.index = kwargs.get('index', None)

        for field_name, field in self._fields.items():
            value = kwargs.get(field_name)
            setattr(self, field_name, value)

        if not self._sheet:
            raise ValueError('Sheet is not bound to the model.')

    async def get_all(self) -> list:
        records = await self._request(
            function=partial(
                self._sheet.get_all_records,
            ),
        )

        instances = []
        for index, record in enumerate(records, start=2):
            instance = self.__class__(index=index)
            for field_name, field in self._fields.items():
                value = record.get(field.title)
                setattr(instance, field_name, value)
            await instance._deserialization()
            instances.append(instance)
        return instances

    async def create(self):
        await self._serialization()
        data = [getattr(self, f'_{field_name}', None) for field_name in self._fields]
        if self.index:
            await self._request(
                function=partial(
                    self._sheet.insert_row,
                    values=data,
                    index=self.index,
                    value_input_option=ValueInputOption.user_entered,
                ),
            )
            return self
        await self._request(
            function=partial(
                self._sheet.append_row,
                values=data,
                value_input_option=ValueInputOption.user_entered,
            ),
        )
        return self

    async def update(self):
        await self._serialization()
        data = [getattr(self, f'_{field_name}', None) for field_name in self._fields]
        await self._request(
            function=partial(
                self._sheet.update,
                values=[data],
                range_name=f'A{self.index}:{chr(64+len(data))}{self.index}',
                value_input_option=ValueInputOption.user_entered,
            ),
        )

    async def _serialization(self):
        for field_name, field in self._fields.items():
            value = getattr(self, field_name, None)

            formatted_value = None

            if not value and field.default:
                value = field.default
                setattr(self, field_name, value)

            if not value and not field.nullable:
                raise TypeError(f'Attribute "{field_name}" cannot be None')

            if value is not None:
                type_ = TYPES[field.type]

                if field.type == FieldType.FLOAT and isinstance(value, int):
                    value = float(value)

                if not isinstance(value, type_):
                    raise TypeError(f'Attribute "{field_name}" must have a type {type_}. Now {type(value)}')

                formatted_value = value
                if field.type in [FieldType.DATETIME, FieldType.DATE, FieldType.TIME]:
                    formatted_value = value.strftime(field.format)

            setattr(self, f'_{field_name}', formatted_value)

    async def _deserialization(self):
        for field_name, field in self._fields.items():
            formatted_value = getattr(self, field_name, None)
            value = None

            if formatted_value:
                if field.type == FieldType.BOOLEAN:
                    value = True if formatted_value == 'TRUE' else False
                elif field.type == FieldType.STRING:
                    value = formatted_value

                # Integer
                elif field.type == FieldType.INTEGER and isinstance(formatted_value, int):
                    value = formatted_value
                elif field.type == FieldType.INTEGER and isinstance(formatted_value, str):
                    if formatted_value.isdigit():
                        value = int(formatted_value)

                elif field.type == FieldType.FLOAT and isinstance(formatted_value, float):
                    value = formatted_value
                elif field.type == FieldType.FLOAT and isinstance(formatted_value, int):
                    value = float(formatted_value)

                # Datetime
                elif field.type == FieldType.DATETIME:
                    value = datetime.strptime(formatted_value, field.format)
                elif field.type == FieldType.DATE:
                    value = datetime.strptime(formatted_value, field.format).date()
                elif field.type == FieldType.TIME:
                    value = datetime.strptime(formatted_value, field.format).time()
                else:
                    raise TypeError(
                        f'Value "{field.title}" on line {self.index} does not match the type "{field.type}". '
                        f'The current type is: "{type(formatted_value)}"'
                    )

            setattr(self, field_name, value)
            setattr(self, f'_{field_name}', formatted_value)

    @staticmethod
    async def _request(function: partial):
        loop = get_running_loop()
        with ThreadPoolExecutor() as pool:
            # noinspection PyTypeChecker
            result = await loop.run_in_executor(pool, function)
        return result

    @classmethod
    def bind_sheet(cls, sheet: Worksheet):
        cls._sheet = sheet



from enum import Enum


class FieldType(str, Enum):
    BOOLEAN = 'BOOLEAN'
    STRING = 'STRING'
    INTEGER = 'INTEGER'
    FLOAT = 'FLOAT'
    DATETIME = 'DATETIME'
    DATE = 'DATE'
    TIME = 'TIME'



from setuptools import setup, find_packages
from os import getenv
# noinspection PyPackageRequirements
from dotenv import load_dotenv


load_dotenv()
NAME = getenv('NAME')
VERSION = getenv('VERSION')
DESCRIPTION = getenv('DESCRIPTION')
AUTHOR = getenv('AUTHOR')
AUTHOR_EMAIL = getenv('AUTHOR_EMAIL')
URL = getenv('URL')


README_REPLACE = {
    '🧡': '♡'
}


def parse_requirements(filename: str) -> list[str]:
    with open(filename, 'r') as f:
        return [line.strip() for line in f if line.strip() and not line.startswith('#')]


def get_long_description() -> str:
    with open('README.md', encoding='utf-8') as file:
        long_description = file.read()
    for key, value in README_REPLACE.items():
        long_description = long_description.replace(key, value)
    return long_description


setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    url=URL,
    packages=find_packages(),
    long_description=get_long_description(),
    long_description_content_type='text/markdown',
    license='MIT',
    install_requires=parse_requirements('requires.txt'),
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.11',
)

